#region Copyright  2009, DevLynx (MIT License)
// This software is licensed under The MIT License:
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#endregion

namespace AeroGlass
{
    using System.Drawing;
    using System.Windows.Forms;


    /// <summary>
    /// Properties associated with the form to be glassed.
    /// </summary>
    public class AeroGlassFormProperties
    {

        /// <summary>
        /// Initializes a new instance of the <see cref="AeroGlassFormProperties"/> class.
        /// </summary>
        public AeroGlassFormProperties()
        {
            GlassEnabled = true;
            GlassTopMovesForm = false;
            GlassMargins = new Padding(0);
        }

        /// <summary>
        /// Gets or sets a value indicating whether the glass effect should be enabled.
        /// </summary>
        /// <value><c>true</c> if glass effect should be enabled; otherwise, <c>false</c>.</value>
        public bool GlassEnabled { get; set; }

        /// <summary>
        /// Gets or sets the glass margins. These margins dictate how much of the client space should be glassed.
        /// If all values are set to -1 the full form will be glassed.
        /// </summary>
        /// <value>The glass margins.</value>
        public Padding GlassMargins { get; set; }

        /// <summary>
        /// <para>Gets or sets a value indicating whether clicking and dragging within the top margin will move the form.
        /// If this is set to true then the form must override it's WinProc and add:
        /// <code>aeroGlassProvider.WndProc(this, ref message);</code>.</para>
        /// </summary>
        /// <value><c>true</c> if clicking and dragging on the top margin moves the form; otherwise, <c>false</c>.</value>
        public bool GlassTopMovesForm { get; set; }

        /// <summary>
        /// Gets a value indicating whether this instance will glass the full form.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if this instance glasses the full form; otherwise, <c>false</c>.
        /// </value>
        public bool IsFullFormGlass
        {
            get { return (GlassMargins.Bottom == -1 && GlassMargins.Top == -1 && GlassMargins.Left == -1 && GlassMargins.Right == -1); }
        }

        /// <summary>
        /// Gets the rectangle associated with the supplied MarginSide.
        /// </summary>
        /// <param name="side">The side of the margin to obtain..</param>
        /// <param name="bounds">The bounds of the form that will be glassed.</param>
        /// <returns>The rectangle associated with the supplied MarginSide.</returns>
        public Rectangle MarginRectangle(MarginSide side, Rectangle bounds)
        {
            Rectangle result;
            if (IsFullFormGlass)
                result = bounds;
            else
            {
                switch (side)
                {
                    case MarginSide.Top:
                        result = new Rectangle(0, 0, bounds.Width, GlassMargins.Top);
                        break;
                    case MarginSide.Left:
                        result = new Rectangle(0, 0, GlassMargins.Left, bounds.Height);
                        break;
                    case MarginSide.Right:
                        result = new Rectangle(bounds.Width - GlassMargins.Right, 0, GlassMargins.Right, bounds.Height);
                        break;
                    case MarginSide.Bottom:
                        result = new Rectangle(0, bounds.Height - GlassMargins.Bottom, bounds.Width, GlassMargins.Bottom);
                        break;
                    default:
                        result = new Rectangle();
                        break;
                }
            }
            return result;
        }
    }
}