#region Copyright  2009, DevLynx (MIT License)
// This software is licensed under The MIT License:
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// Algorithms and inspiration from:
//   http://channel9.msdn.com/playground/Sandbox/201158/
//   http://www.codeproject.com/KB/vista/textonglass.aspx
#endregion

namespace AeroGlass
{
    using System;
    using System.Diagnostics;
    using System.Drawing;
    using System.Windows.Forms;
    using System.Windows.Forms.VisualStyles;

    /// <summary>
	/// This utility extends the Vista's glass effect into a windows form and draws text with glow onto the glass.
	/// </summary>
	public static class AeroGlassUtility
	{

		/// <summary>
		/// Returns whether the Aero interface is enabled.
		/// </summary>
		public static bool IsAeroEnabled
		{
			get
			{
                return IsAeroSupported && IsCompositionEnabled();
			}
		}

		/// <summary>
		/// Returns whether the Aero interface is supported by the OS.
		/// </summary>
		public static bool IsAeroSupported
		{
            get { return Environment.OSVersion.Version.Major >= 6; }
		}


        /// <summary>
        /// Draws the PictureBox contents onto the glass.
        /// </summary>
        /// <param name="form">The form where the glass is located.</param>
        /// <param name="image">The image to be drawn.</param>
        /// <param name="rectangle">The rectangle in which to draw the image.</param>
        public static void DrawImageOnGlass(Form form, Image image, Rectangle rectangle)
        {
            using (Graphics grapicSurface = form.CreateGraphics())
                grapicSurface.DrawImage(image, rectangle);
        }

        /// <summary>
        /// Draws the text on glass with a glow effect behind the text.
        /// </summary>
        /// <param name="form">The form to draw the glass.</param>
        /// <param name="text">The text to draw.</param>
        /// <param name="font">The font in which to draw the text.</param>
        /// <param name="rectangle">The rectangle in which to draw the text.</param>
        /// <param name="glowSize">Size of the glow behind the text.</param>
        public static void DrawTextOnGlass(Form form, string text, Font font, Rectangle rectangle, int glowSize)
        {
            if (IsAeroSupported && IsAeroEnabled)
            {
                IntPtr destdc = InteropServices.GetDC(form.Handle);
                IntPtr Memdc = InteropServices.CreateCompatibleDC(destdc);
                if (!(InteropServices.SaveDC(Memdc) == 0))
                {
                    int doubledGlowSize = 2 * glowSize;
                    Rectangle xRectWithGlow = new Rectangle(rectangle.Left, rectangle.Top,
                        rectangle.Width + doubledGlowSize, rectangle.Height + doubledGlowSize);
                    InteropServices.BITMAPINFO dib = new InteropServices.BITMAPINFO(xRectWithGlow.Width, -xRectWithGlow.Height);
                    IntPtr bitmap = InteropServices.CreateDIBSection(Memdc, ref dib, InteropServices.DIB_RGB_COLORS, 0, IntPtr.Zero, 0);
                    if ((bitmap != IntPtr.Zero))
                    {
                        IntPtr bitmapOld = IntPtr.Zero;
                        bitmapOld = InteropServices.SelectObject(Memdc, bitmap);
                        IntPtr hFont = font.ToHfont();
                        IntPtr logfnotOld;
                        logfnotOld = InteropServices.SelectObject(Memdc, hFont);
                        try
                        {
                            VisualStyleRenderer renderer = new VisualStyleRenderer(VisualStyleElement.Window.Caption.Active);
                            Rectangle xNormalizedRectWithGlow = new Rectangle(0, 0, xRectWithGlow.Width, xRectWithGlow.Height);
                            InteropServices.RECT normalizedRectWithGlow = new InteropServices.RECT(xNormalizedRectWithGlow);
                            InteropServices.DTTOPTS dttOpts = new InteropServices.DTTOPTS(glowSize);
                            int uFormat = InteropServices.DT_SINGLELINE | InteropServices.DT_CENTER | InteropServices.DT_VCENTER | InteropServices.DT_NOPREFIX;
                            InteropServices.DrawThemeTextEx(renderer.Handle, Memdc, 0, 0, text, -1, uFormat, ref normalizedRectWithGlow, ref dttOpts);
                            InteropServices.BitBlt(destdc, xRectWithGlow.Left - glowSize, xRectWithGlow.Top - glowSize,
                                xRectWithGlow.Width, xRectWithGlow.Height, Memdc, 0, 0, InteropServices.SRCCOPY);
                        }
                        catch (Exception e)
                        {
                            Trace.WriteLine(e.Message);
                        }
                        finally
                        {
                            InteropServices.SelectObject(Memdc, bitmapOld);
                            InteropServices.SelectObject(Memdc, logfnotOld);
                            InteropServices.DeleteObject(bitmap);
                            InteropServices.DeleteObject(hFont);

                            InteropServices.ReleaseDC(Memdc, -1);
                            InteropServices.DeleteDC(Memdc);
                        }
                    }
                }
            }
        }

        private static bool IsCompositionEnabled()
        {
            return InteropServices.DwmIsCompositionEnabled();
        }

        internal static IntPtr CaptionIntPtr()
        {
            return new IntPtr(InteropServices.HTCAPTION);
        }

        internal static bool CompositionChanged(Message message)
        {
            return InteropServices.DwmCompositionChanged(message);
        }

        internal static bool CompositionColorChanged(Message message)
        {
            return InteropServices.DwmCompositionColorChanged(message);
        }

        internal static void GetAeroColor(out int glassColor, out bool isOpaque)
        {
            InteropServices.DwmGetColorizationColor(out glassColor, out isOpaque);
        }

        internal static bool MouseClickOnClient(Message message)
        {
            return message.Msg == InteropServices.WM_NCHITTEST && message.Result.ToInt32() == InteropServices.HTCLIENT;
        }

        internal static void PaintGlassEffect(Form form, Padding margins)
        {
            InteropServices.MARGINS nativeMargins = new InteropServices.MARGINS(margins);
            InteropServices.DwmExtendFrameIntoClientArea(form.Handle, ref nativeMargins);
        }

        internal static Point ScreenCoordinates(Message message)
        {
            // Get the screen coordinates.
            int lParam = message.LParam.ToInt32();
            return new Point((lParam << 16) >> 16, lParam >> 16);
        }
    }
}
